import subprocess

import jgo.util
import os
import pathlib
import sys

from paintera import version
from paintera.javafx_modules import javafx_args

_paintera                = '@Paintera'
_paintera_show_container = '@PainteraShowContainer'
_picocli_autocomplete    = 'picocli.AutoComplete'
_paintera_cli_args       = 'org.janelia.saalfeldlab.paintera.PainteraCommandLineArgs'
_groupId                 = 'org.janelia.saalfeldlab'
_artifactId              = 'paintera'

def _get_paintera_version(argv=None):
    import argparse
    argv = sys.argv[1:] if argv is None else argv
    if ('--' in argv):

        def version_from_string(string):
            split = string.split('.')
            try:
                major, minor, patch = [int(s) for s in split[:3]]
                tag = ''
            except:
                major, minor = [int(s) for s in split[:2]]
                patch = int(split[2].split('-')[0])
                tag = 'SNAPSHOT'
            return version._Version(major, minor, patch, tag)

        parser = argparse.ArgumentParser(usage=argparse.SUPPRESS)
        parser.add_argument('--use-version', type=version_from_string)
        args, unknown = parser.parse_known_args(argv)
        return version._paintera_version if args.use_version is None else args.use_version, unknown

    else:
        return version._paintera_version, argv


def prepend_required_default_args(argv):
    # Start with the manual scaling args, to support HiDPI scaling on linus
    additional_args = manual_scaling_args()
    # Add args to ensure GPU is always used (and required)
    additional_args.append("-Dprism.forceGPU=true")
    # Add the javafx arguments, if not previously set
    if "--module-path" not in argv:
        additional_args.extend(javafx_args)
    # Append the arg separator if not present.
    if '--' not in argv:
        additional_args.append('--')
    return additional_args + argv

def launch_paintera():
    paintera_version, argv = _get_paintera_version(argv=sys.argv[1:])
    final_args = prepend_required_default_args(argv)
    return jgo.util.main_from_endpoint(
        argv=final_args,
        primary_endpoint=f'{_groupId}:{_artifactId}',
        primary_endpoint_version=paintera_version.maven_version(),
        primary_endpoint_main_class=_paintera)


def manual_scaling_args():
    """
    This determines and provides the appropriate scaling properties for javafx on hi-dpi monitors, when running linux.
    This will only impact scaling when using Mutter (default for GNOME 3 )

    :return: list additional javafx properties which override application scaling
    """
    if 'linux' in sys.platform:
        scale = dbus_request_mutter_screen_scale_factor()
        if (scale == None or len(scale) == 0):
            #  If there request doesn't work, don't attempt to override scaling. Paintera should work,
            #   though the resolution may not match the system's display configuration.
            return []
        return ["-Dprism.allowhidpi=true", f"-Dglass.gtk.uiScale={scale}"]
    else:
        return []


def dbus_request_mutter_screen_scale_factor():
    """
    Queries the Mutter DisplayConfig API via dbus to determine the current scale factor.

    :return: the current scale factor used by the Mutter display manager on linux
    """
    dbus_name = 'org.gnome.Mutter.DisplayConfig'
    message_name = f"{dbus_name}.GetCurrentState"
    dest_obj_path = '/org/gnome/Mutter/DisplayConfig'
    return subprocess.run(
        f"dbus-send --print-reply --dest={dbus_name} {dest_obj_path} {message_name} | grep -i scaling-factor -A 1 | grep int32 " + "| awk '{print $3}'",
        capture_output=True, text=True, shell=True).stdout.strip()


def generate_paintera_bash_completion():
    relative_path = os.path.join(
        'share',
        'bash-completion',
        'completions',
        'paintera')
    default_path = os.path.join('$PREFIX', relative_path)
    description=f'''Generate bash completion file for Paintera. This script uses picocli[1] to create a script file for bash-completion[2]. The default output location is {default_path} and can be modified with the --prefix option or overwritten entirely with the --output option.

[1] https://picocli.info
[2] https://github.com/scop/bash-completion
'''
    import argparse
    parser = argparse.ArgumentParser(
        description=description,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    default_prefix = os.getenv('PREFIX', os.path.abspath(os.path.join(os.sep, 'usr')))
    parser.add_argument('--prefix', required=False, default=default_prefix, help='Sets the PREFIX for the output location. Is ignored if --output is set.')
    parser.add_argument('--output', required=False, default=None, help='Specifies the output location explicitly (PREFIX and --prefix are ignored)')

    argv = sys.argv[1:]

    try:
        double_dash_index = argv.index('--')
    except ValueError:
        double_dash_index = -1

    if double_dash_index > 0 and double_dash_index < len(argv):
        jgo_argv = argv[:double_dash_index + 1]
        argv     = argv[double_dash_index + 1:]
    else:
        jgo_argv = []

    args = parser.parse_args(args=argv)

    output = os.path.join(args.prefix, relative_path) if args.output is None else args.output

    argv = [
        '-n', 'paintera',
        'org.janelia.saalfeldlab.paintera.PainteraCommandLineArgs',
        '-o', output]

    pathlib.Path(output).parent.mkdir(parents=True, exist_ok=True)

    return jgo.util.main_from_endpoint(
        primary_endpoint            = f'{_groupId}:{_artifactId}',
        argv                        = jgo_argv + argv,
        primary_endpoint_version    = version._paintera_version.maven_version(),
        primary_endpoint_main_class = _picocli_autocomplete)

if __name__ == '__main__':
    launch_paintera()


import os
import pathlib
import sys

from jgo import jgo

from paintera import version

_paintera_endpoint = f'org.janelia.saalfeldlab:paintera:{version._paintera_version.maven_version()}'

_classpath_separator = ':' if sys.platform != 'win32' else ';'
_javafx_platform_map = {
    "linux": "linux",
    "linux2": "linux",
    "win32": "win",
    "darwin": "mac"
}

_javafx_version = "16"
_modules_and_opens = {
    "javafx.base": [
        'javafx.util',
        'javafx.event',
        'javafx.beans.property',
        'com.sun.javafx.binding',
        'com.sun.javafx.event',
    ],
    "javafx.controls": [],
    "javafx.fxml": [],
    "javafx.media": [],
    "javafx.swing": [],
    "javafx.web": [],
    "javafx.graphics": [
        "javafx.scene",
        "javafx.stage",
        "javafx.geometry",
        "javafx.animation",
        "javafx.scene.input",
        "javafx.scene.image",
        "com.sun.prism",
        "com.sun.prism.paint",
        "com.sun.javafx.application",
        "com.sun.javafx.geom",
        "com.sun.javafx.image",
        "com.sun.javafx.scene",
        "com.sun.javafx.stage",
        "com.sun.javafx.perf",
        "com.sun.javafx.cursor",
        "com.sun.javafx.tk",
        "com.sun.javafx.scene.traversal",
        "com.sun.javafx.geom.transform",
        "com.sun.scenario.animation",
        "com.sun.scenario.animation.shared",
        "com.sun.scenario.effect",
        "com.sun.javafx.sg.prism",
    ]
}


def _get_jgo_cache_dir():
    config = jgo.default_config()
    if not '--ignore-jgorc' in sys.argv:
        config_file = pathlib.Path.home() / '.jgorc'
        config.read(config_file)

    if os.getenv(jgo.jgo_cache_dir_environment_variable()) is not None:
        cache_dir = os.getenv(jgo.jgo_cache_dir_environment_variable())
        config.set('settings', 'cacheDir', cache_dir)
    return config['settings'].get('cacheDir')


def _javafx_module_path():
    """
    We need to determine where jgo copies the jar dependencies to, then build out module-path from that.
    We grab the endpoint we pass to jgo, then call jgo's internal methods for determining the workspace.
    After that, we determine the javafx dependencies we need to add the module-path. For each dependency
     there is a "common" jar, as a well as a platform specific jar.

    :return: the module-path which points to the javafx module dependencies
    """
    # Determine the location of the jgo-discovered dependencies
    endpoints = jgo.endpoints_from_strings([_paintera_endpoint])
    coordinates = jgo.coordinates_from_endpoints(endpoints)
    cache_dir = _get_jgo_cache_dir()
    workspace = jgo.workspace_dir_from_coordinates(coordinates, cache_dir=cache_dir)
    relative_module_paths = []
    for module in _modules_and_opens.keys():
        # generate a list of the dependencies we should expect to find.
        module_path_name = module.replace(".", "-")
        shared_prefix = f"{module_path_name}-{_javafx_version}"
        common_module = f"{shared_prefix}.jar"
        platform_specific_module = f"{shared_prefix}-{_javafx_platform_map[sys.platform]}.jar"
        relative_module_paths.append(platform_specific_module),
        relative_module_paths.append(common_module)
    # create the module path
    absolute_module_paths = (f"{workspace}/{x}" for x in relative_module_paths)
    module_path = _classpath_separator.join(absolute_module_paths)
    return module_path


# Allow access to classes that are not exported (currently needed)
javafx_args = ["--illegal-access=permit"]

# for each class we use, open it to ALL-UNNAMED, which contains Paintera,
#   since it isn't utilizing the module system currently
for module_name, packages_to_open in _modules_and_opens.items():
    for package in packages_to_open:
        javafx_args.append(f"--add-opens={module_name}/{package}=ALL-UNNAMED")

# add the module path for the javafx modules
javafx_args.extend(["--module-path", _javafx_module_path()])
# specify which modules we actually use.
javafx_args.extend(["--add-modules", ",".join(_modules_and_opens.keys())])
